@@ -85,9 +85,6 @@ editor_source_view_update_css (EditorSourceView *self)
   const PangoFontDescription *font_desc;
   PangoFontDescription *scaled = NULL;
   PangoFontDescription *system_font = NULL;
-  GtkSourceStyleScheme *scheme;
-  GtkSourceStyle *style;
-  GtkTextBuffer *buffer;
   g_autoptr(GString) str = NULL;
   g_autofree char *font_css = NULL;
   int size = 11; /* 11pt */
@@ -97,29 +94,6 @@ editor_source_view_update_css (EditorSourceView *self)
 
   str = g_string_new (NULL);
 
-  /* Get information for search bubbles */
-  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (self));
-  if ((scheme = gtk_source_buffer_get_style_scheme (GTK_SOURCE_BUFFER (buffer))) &&
-      (style = gtk_source_style_scheme_get_style (scheme, "search-match")))
-    {
-      g_autofree char *background = NULL;
-      gboolean background_set = FALSE;
-
-      g_object_get (style,
-                    "background", &background,
-                    "background-set", &background_set,
-                    NULL);
-
-      if (background != NULL && background_set)
-        g_string_append_printf (str,
-                                ".search-match {"
-                                " background:mix(%s,currentColor,0.0125);"
-                                " border-radius:7px;"
-                                " box-shadow: 0 1px 3px mix(%s,currentColor,.2);"
-                                "}\n",
-                                background, background);
-    }
-
   g_string_append (str, "textview {\n");
 
   /* Get font information to adjust line height and font changes */
@@ -338,6 +312,33 @@ on_scroll_scrolled_cb (GtkEventControllerScroll *scroll,
   return FALSE;
 }
 
+static void
+on_scroll_begin_cb (GtkEventControllerScroll *scroll,
+                    EditorSourceView         *self)
+{
+  GdkModifierType state;
+
+  g_assert (GTK_IS_EVENT_CONTROLLER_SCROLL (scroll));
+  g_assert (EDITOR_IS_SOURCE_VIEW (self));
+
+  state = gtk_event_controller_get_current_event_state (GTK_EVENT_CONTROLLER (scroll));
+
+  if (state & GDK_CONTROL_MASK)
+    gtk_event_controller_scroll_set_flags (scroll,
+                                           GTK_EVENT_CONTROLLER_SCROLL_VERTICAL |
+                                           GTK_EVENT_CONTROLLER_SCROLL_DISCRETE);
+}
+
+static void
+on_scroll_end_cb (GtkEventControllerScroll *scroll,
+                  EditorSourceView         *self)
+{
+  g_assert (GTK_IS_EVENT_CONTROLLER_SCROLL (scroll));
+  g_assert (EDITOR_IS_SOURCE_VIEW (self));
+
+  gtk_event_controller_scroll_set_flags (scroll, GTK_EVENT_CONTROLLER_SCROLL_VERTICAL);
+}
+
 static void
 on_notify_buffer_cb (EditorSourceView *self,
                      GParamSpec       *pspec,
@@ -461,6 +462,29 @@ editor_source_view_action_zoom (GtkWidget  *widget,
     g_assert_not_reached ();
 }
 
+static void
+select_line (GtkTextBuffer *buffer,
+             GtkTextIter   *begin,
+             GtkTextIter   *end)
+{
+  g_assert (GTK_IS_TEXT_BUFFER (buffer));
+
+  gtk_text_buffer_get_selection_bounds (buffer, begin, end);
+  gtk_text_iter_order (begin, end);
+
+  /* Move beginning of selection/cursor to position 0 of first */
+  if (!gtk_text_iter_starts_line (begin))
+    gtk_text_iter_set_line_offset (begin, 0);
+
+  /* Move end of selection/cursor to end of line */
+  if (!gtk_text_iter_ends_line (end))
+    gtk_text_iter_forward_to_line_end (end);
+
+  /* Swallow the \n with the selection */
+  if (!gtk_text_iter_is_end (end))
+    gtk_text_iter_forward_char (end);
+}
+
 static void
 editor_source_view_action_select_line (GtkWidget  *widget,
                                        const char *action_name,
@@ -473,21 +497,7 @@ editor_source_view_action_select_line (GtkWidget  *widget,
   g_assert (EDITOR_IS_SOURCE_VIEW (self));
 
   buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (self));
-  gtk_text_buffer_get_selection_bounds (buffer, &begin, &end);
-  gtk_text_iter_order (&begin, &end);
-
-  /* Move beginning of selection/cursor to position 0 of first */
-  if (!gtk_text_iter_starts_line (&begin))
-    gtk_text_iter_set_line_offset (&begin, 0);
-
-  /* Move end of selection/cursor to end of line */
-  if (!gtk_text_iter_ends_line (&end))
-    gtk_text_iter_forward_to_line_end (&end);
-
-  /* Swallow the \n with the selection */
-  if (!gtk_text_iter_is_end (&end))
-    gtk_text_iter_forward_char (&end);
-
+  select_line (buffer, &begin, &end);
   gtk_text_buffer_select_range (buffer, &begin, &end);
 
   /* NOTE: This shouldn't be needed, but due to some invalidation issues in
@@ -509,7 +519,6 @@ editor_source_view_action_delete_line (GtkWidget  *widget,
 
   g_assert (EDITOR_IS_SOURCE_VIEW (self));
 
-  editor_source_view_action_select_line (widget, NULL, NULL);
   buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (self));
 
   /* If we're at the end of the buffer, then we need to remove the
@@ -517,7 +526,7 @@ editor_source_view_action_delete_line (GtkWidget  *widget,
    * user that a line was removed.
    */
   gtk_text_buffer_begin_user_action (buffer);
-  gtk_text_buffer_get_selection_bounds (buffer, &begin, &end);
+  select_line (buffer, &begin, &end);
   gtk_text_iter_order (&begin, &end);
   if (gtk_text_iter_is_end (&end) &&
       gtk_text_iter_get_line (&begin) == gtk_text_iter_get_line (&end))
@@ -539,179 +548,51 @@ editor_source_view_action_delete_line (GtkWidget  *widget,
   gdk_clipboard_set_text (gtk_widget_get_primary_clipboard (widget), text);
 }
 
-static GtkSourceSearchContext *
-get_search_context (EditorSourceView *self)
-{
-  EditorPage *page = EDITOR_PAGE (gtk_widget_get_ancestor (GTK_WIDGET (self), EDITOR_TYPE_PAGE));
-
-  if (page->search_bar->context &&
-      gtk_source_search_context_get_highlight (page->search_bar->context))
-    return page->search_bar->context;
-
-  return NULL;
-}
-
-static inline void
-add_match (GtkTextView       *text_view,
-           cairo_region_t    *region,
-           const GtkTextIter *begin,
-           const GtkTextIter *end)
-{
-  GdkRectangle begin_rect;
-  GdkRectangle end_rect;
-  cairo_rectangle_int_t rect;
-
-  /* NOTE: @end is not inclusive of the match. */
-  if (gtk_text_iter_get_line (begin) == gtk_text_iter_get_line (end))
-    {
-      gtk_text_view_get_iter_location (text_view, begin, &begin_rect);
-      gtk_text_view_get_iter_location (text_view, end, &end_rect);
-      rect.x = begin_rect.x;
-      rect.y = begin_rect.y;
-      rect.width = end_rect.x - begin_rect.x;
-      rect.height = MAX (begin_rect.height, end_rect.height);
-
-      cairo_region_union_rectangle (region, &rect);
-
-      return;
-    }
-
-  /*
-   * TODO: Add support for multi-line matches. When @begin and @end are not
-   *       on the same line, we need to add the match region to @region so
-   *       ide_source_view_draw_search_bubbles() can draw search bubbles
-   *       around it.
-   */
-}
-
-static guint
-add_matches (GtkTextView            *text_view,
-             cairo_region_t         *region,
-             GtkSourceSearchContext *search_context,
-             const GtkTextIter      *begin,
-             const GtkTextIter      *end)
-{
-  GtkTextIter first_begin;
-  GtkTextIter new_begin;
-  GtkTextIter match_begin;
-  GtkTextIter match_end;
-  gboolean has_wrapped;
-  guint count = 1;
-
-  g_assert (GTK_IS_TEXT_VIEW (text_view));
-  g_assert (region != NULL);
-  g_assert (GTK_SOURCE_IS_SEARCH_CONTEXT (search_context));
-  g_assert (begin != NULL);
-  g_assert (end != NULL);
-
-  if (!gtk_source_search_context_forward (search_context,
-                                          begin,
-                                          &first_begin,
-                                          &match_end,
-                                          &has_wrapped))
-    return 0;
-
-  add_match (text_view, region, &first_begin, &match_end);
-
-  for (;;)
-    {
-      new_begin = match_end;
-
-      if (gtk_source_search_context_forward (search_context,
-                                             &new_begin,
-                                             &match_begin,
-                                             &match_end,
-                                             &has_wrapped) &&
-          (gtk_text_iter_compare (&match_begin, end) < 0) &&
-          (gtk_text_iter_compare (&first_begin, &match_begin) != 0))
-        {
-          add_match (text_view, region, &match_begin, &match_end);
-          count++;
-          continue;
-        }
-
-      break;
-    }
-
-  return count;
-}
-
-G_GNUC_UNUSED static void
-editor_source_view_draw_search_bubbles (EditorSourceView *self,
-                                        GtkSnapshot      *snapshot)
-{
-  GtkStyleContext *style_context;
-  cairo_region_t *clip_region;
-  cairo_region_t *match_region;
-  GtkSourceSearchContext *context;
-  GdkRectangle area;
-  GtkTextIter begin, end;
-
-  g_assert (EDITOR_IS_SOURCE_VIEW (self));
-  g_assert (GTK_IS_SNAPSHOT (snapshot));
-
-  if (self->font_scale < MIN_BUBBLE_SCALE ||
-      self->font_scale > MAX_BUBBLE_SCALE ||
-      !(context = get_search_context (self)))
-    return;
-
-  style_context = gtk_widget_get_style_context (GTK_WIDGET (self));
-
-  gtk_text_view_get_visible_rect (GTK_TEXT_VIEW (self), &area);
-  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (self), &begin,
-                                      area.x, area.y);
-  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (self), &end,
-                                      area.x + area.width,
-                                      area.y + area.height);
-
-  clip_region = cairo_region_create_rectangle (&area);
-  match_region = cairo_region_create ();
-
-  if (add_matches (GTK_TEXT_VIEW (self), match_region, context, &begin, &end))
-    {
-      int n_rects;
-
-      cairo_region_subtract (clip_region, match_region);
-
-      n_rects = cairo_region_num_rectangles (match_region);
-
-      gtk_style_context_save (style_context);
-      gtk_style_context_add_class (style_context, "search-match");
-      for (guint i = 0; i < n_rects; i++)
-        {
-          cairo_rectangle_int_t r;
-
-          cairo_region_get_rectangle (match_region, i, &r);
-
-          r.x -= X_PAD;
-          r.width += 2*X_PAD;
-          r.y -= Y_PAD;
-          r.height += 2*Y_PAD;
-
-          gtk_snapshot_render_background (snapshot, style_context, r.x, r.y, r.width, r.height);
-          gtk_snapshot_render_frame (snapshot, style_context, r.x, r.y, r.width, r.height);
-        }
-      gtk_style_context_restore (style_context);
-    }
-
-  cairo_region_destroy (clip_region);
-  cairo_region_destroy (match_region);
-}
-
 static void
-editor_source_view_snapshot_layer (GtkTextView      *text_view,
-                                   GtkTextViewLayer  layer,
-                                   GtkSnapshot      *snapshot)
+editor_source_view_action_duplicate_line (GtkWidget  *widget,
+                                          const char *action_name,
+                                          GVariant   *param)
 {
-  EditorSourceView *self = (EditorSourceView *)text_view;
+  g_autofree char *text = NULL;
+  g_autofree char *duplicate_line = NULL;
+  GtkTextIter begin, end;
+  GtkTextMark *cursor;
+  GtkTextBuffer *buffer;
+  gboolean selected;
 
-  g_assert (EDITOR_IS_SOURCE_VIEW (self));
-  g_assert (GTK_IS_SNAPSHOT (snapshot));
+  g_assert (EDITOR_IS_SOURCE_VIEW (widget));
 
-  GTK_TEXT_VIEW_CLASS (editor_source_view_parent_class)->snapshot_layer (text_view, layer, snapshot);
+  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widget));
+  cursor = gtk_text_buffer_get_insert (buffer);
 
-  if (layer == GTK_TEXT_VIEW_LAYER_BELOW_TEXT)
-    editor_source_view_draw_search_bubbles (self, snapshot);
+  gtk_text_buffer_begin_user_action (buffer);
+
+  selected = gtk_text_buffer_get_selection_bounds (buffer, &begin, &end);
+
+  if (selected)
+    {
+      duplicate_line = gtk_text_iter_get_text (&begin, &end);
+      gtk_text_buffer_insert (buffer, &begin, duplicate_line, -1);
+    }
+  else
+    {
+      gtk_text_buffer_get_iter_at_mark (buffer, &begin, cursor);
+      end = begin;
+
+      gtk_text_iter_set_line_offset (&begin, 0);
+
+      if (!gtk_text_iter_ends_line (&end))
+        gtk_text_iter_forward_to_line_end (&end);
+
+      if (gtk_text_iter_get_line (&begin) == gtk_text_iter_get_line (&end))
+        {
+          text = gtk_text_iter_get_text (&begin, &end);
+          duplicate_line = g_strconcat (text, "\n", NULL);
+          gtk_text_buffer_insert (buffer, &begin, duplicate_line, -1);
+        }
+    }
+
+  gtk_text_buffer_end_user_action (buffer);
 }
 
 static gboolean
@@ -857,7 +738,6 @@ editor_source_view_class_init (EditorSourceViewClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
-  GtkTextViewClass *text_view_class = GTK_TEXT_VIEW_CLASS (klass);
 
   object_class->constructed = editor_source_view_constructed;
   object_class->dispose = editor_source_view_dispose;
@@ -867,8 +747,6 @@ editor_source_view_class_init (EditorSourceViewClass *klass)
   widget_class->root = editor_source_view_root;
   widget_class->size_allocate = editor_source_view_size_allocate;
 
-  text_view_class->snapshot_layer = editor_source_view_snapshot_layer;
-
   properties [PROP_LINE_HEIGHT] =
     g_param_spec_double ("line-height",
                          "Line height",
@@ -907,6 +785,7 @@ editor_source_view_class_init (EditorSourceViewClass *klass)
   gtk_widget_class_install_action (widget_class, "spelling.correct", "s", editor_source_view_action_spelling_correct);
   gtk_widget_class_install_action (widget_class, "buffer.select-line", NULL, editor_source_view_action_select_line);
   gtk_widget_class_install_action (widget_class, "buffer.delete-line", NULL, editor_source_view_action_delete_line);
+  gtk_widget_class_install_action (widget_class, "buffer.duplicate-line", NULL, editor_source_view_action_duplicate_line);
 
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_plus, GDK_CONTROL_MASK, "page.zoom-in", NULL);
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_KP_Add, GDK_CONTROL_MASK, "page.zoom-in", NULL);
@@ -916,6 +795,7 @@ editor_source_view_class_init (EditorSourceViewClass *klass)
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_0, GDK_CONTROL_MASK, "page.zoom-one", NULL);
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_l, GDK_CONTROL_MASK, "buffer.select-line", NULL);
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_d, GDK_CONTROL_MASK, "buffer.delete-line", NULL);
+  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_d, GDK_CONTROL_MASK|GDK_ALT_MASK, "buffer.duplicate-line", NULL);
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_g, GDK_CONTROL_MASK, "search.move-next", "b", FALSE);
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_g, GDK_CONTROL_MASK|GDK_SHIFT_MASK, "search.move-previous", "b", FALSE);
 }
@@ -971,6 +851,14 @@ editor_source_view_init (EditorSourceView *self)
                     "scroll",
                     G_CALLBACK (on_scroll_scrolled_cb),
                     self);
+  g_signal_connect (controller,
+                    "scroll-begin",
+                    G_CALLBACK (on_scroll_begin_cb),
+                    self);
+  g_signal_connect (controller,
+                    "scroll-end",
+                    G_CALLBACK (on_scroll_end_cb),
+                    self);
   gtk_widget_add_controller (GTK_WIDGET (self), controller);
 
   tweak_gutter_spacing (GTK_SOURCE_VIEW (self));
